from pprint import pformat

from .path import parse_text, NotPresent


class ConfigNode(object):
    """
    A node in the configuration store.
    These are obtained by using the methods below on :class:`~configurator.Config`
    objects or any :class:`ConfigNode` objects returned by them.
    """

    __slots__ = ('_container', '_accessor', 'data')

    def __init__(self, data=None, container=None, accessor=None):
        if data is None:
            data = {}
        #: The underlying python object for this node, often a :class:`dict`
        #: or a :class:`list`. While this is part of the public API, care should
        #: be taken when using it, particularly if you choose to modify it.
        self.data = data
        self._container = container
        self._accessor = accessor

    def _wrap(self, accessor, value):
        if isinstance(value, (dict, list)):
            value = ConfigNode(value, self.data, accessor)
        return value

    def _get(self, name):
        try:
            value = self.data[name]
        except TypeError:
            raise KeyError(name)
        return self._wrap(name, value)

    def __getattr__(self, name):
        """
        Obtain a child of this node by attribute access. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        try:
            return self._get(name)
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """
        Set a child of this node. This is a convenience helper that
        calls :meth:`__setitem__` and can be used when ``name`` is a string.
        """
        if name in self.__slots__:
            super(ConfigNode, self).__setattr__(name, value)
        else:
            self[name] = value

    def __getitem__(self, name):
        """
        Obtain a child of this node by item access. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        return self._get(name)

    def __setitem__(self, name, value):
        """
        Set the ``value`` for the supplied ``name`` in :attr:`data`.
        If :attr:`data` is a :class:`dict`, then ``name`` must be a :class:`str`.
        If :attr:`data` is a :class:`list`, then ``name`` must be an :class:`int`.
        """
        self.data[name] = value

    def get(self, name, default=None):
        """
        Obtain a child of this node by access like :meth:`dict.get`. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        try:
            return self._get(name)
        except KeyError:
            return default

    def items(self):
        """
        Obtain children of this node by access like :meth:`dict.items`. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        for key, value in sorted(self.data.items()):
            yield key, self._wrap(key, value)

    def __iter__(self):
        """
        Obtain children of this node by either :class:`dict` or :class:`list`
        iteration, depending on the type of the underlying :attr:`data`.
        If the child is a :class:`dict` or :class:`list`, a :class:`ConfigNode`
        for it will be returned, otherwise the value itself will be returned.
        """
        for index, item in enumerate(self.data):
            yield self._wrap(index, item)

    def node(self, path=None, create=False):
        """
        Obtain a child of this node using a dotted path or
        :class:`~configurator.path.Path` such as one generated from
        :attr:`~configurator.source`. ``path`` may also be a simple string or integer, in
        which case it will be used to obtain a child of this node using item access.

        This always returns a :class:`ConfigNode` or raises an exception if the path
        cannot be resolved. This allows you to use :meth:`set` even for values in dictionaries
        or items in lists.

        If ``create`` is ``True``, all nodes along the path will be created as dictionaries
        if they do not exist.
        """
        if path is None:
            return self

        path = parse_text(path)
        if not path.ops:
            return self

        if create:
            action = 'ensure'
        else:
            action = 'get'

        data = self.data
        for op in path.ops:
            if isinstance(data, NotPresent):
                op.not_present(data)
            else:
                container = data
                data = getattr(op, action)(container)

        if isinstance(data, NotPresent):
            raise data

        text = getattr(op, 'text', None)
        if text is None:
            raise TypeError('invalid path: '+str(path))

        return ConfigNode(data, container, op.text)

    def set(self, value):
        """
        Replace the :attr:`data` of this node with the supplied ``value``.
        """
        if self._container is None:
            self.data = value
        else:
            self._container[self._accessor] = value

    def __repr__(self):
        cls = type(self)
        pretty = pformat(self.data, width=70)
        if '\n' in pretty:
            pretty = '\n'+pretty+'\n'
        return '{}.{}({})'.format(
            cls.__module__,
            cls.__name__,
            pretty
        )



from collections import defaultdict


class ParseError(Exception):
    """
    The exception raised when an appropriate parser cannot be found for a config
    stream.
    """


class Parsers(defaultdict):

    supported = {
        'json': ('json', 'load'),
        'toml': ('toml', 'load'),
        'yml': ('yaml', 'safe_load'),
        'yaml': ('yaml', 'safe_load'),
    }

    def __missing__(self, extension):
        try:
            module_name, parser_name = self.supported[extension]
        except KeyError:
            raise ParseError('No parser found for {!r}'.format(extension))
        else:
            module = __import__(module_name)
            return getattr(module, parser_name)





def merge_dict(context, source, target):
    result = target.copy()
    for key, source_value in source.items():
        if key in result:
            target_value = result[key]
        else:
            target_value = type(source_value)()
        try:
            value = context.merge(source_value, target_value)
        except TypeError:
            # can't merge, so overwrite
            value = source_value
        result[key] = value
    return result


def merge_list(context, source, target):
    return target + source


class MergeableDict(dict):

    def __add__(self, other):
        result = self.copy()
        result.update(other)
        return result


default_mergers = MergeableDict({
    dict: merge_dict,
    list: merge_list,
})


class MergeContext(object):

    mergers = default_mergers

    def __init__(self, mergers=None):
        if mergers is not None:
            self.mergers = mergers

    def merge(self, source, target):
        source_type = type(source)
        target_type = type(target)
        merger = self.mergers.get(target_type)
        if source_type is not target_type or merger is None:
            raise TypeError('Cannot merge {} with {}'.format(
                source_type, target_type
            ))
        return merger(self, source, target)



# -*- coding: utf-8 -*-
import datetime, os, pkg_resources

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

intersphinx_mapping = {'http://docs.python.org': None}

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx'
    ]

# General
source_suffix = '.rst'
master_doc = 'index'
project = 'configurator'
copyright = '2011-2014 Simplistix Ltd, 2016-%s Chris Withers' % datetime.datetime.now().year
version = release = pkg_resources.get_distribution(project).version
exclude_patterns = [
    'description.rst',
    '_build'
]
pygments_style = 'sphinx'

# Options for HTML output
html_theme = 'default' if on_rtd else 'sphinx_rtd_theme'
htmlhelp_basename = project+'doc'

# Options for LaTeX output
latex_documents = [
  ('index',project+'.tex', project+u' Documentation',
   'Chris Withers', 'manual'),
]




from textwrap import dedent

from testfixtures import compare, ShouldRaise

from configurator import source, convert
from configurator.node import ConfigNode
from configurator.path import NotPresent


class TestInstantiation(object):

    def test_empty(self):
        config = ConfigNode()
        compare(config.data, expected={})

    def test_dict(self):
        config = ConfigNode(dict(x=1))
        compare(config['x'], expected=1)

    def test_list(self):
        config = ConfigNode([1, 2])
        compare(config[0], expected=1)
        compare(config[1], expected=2)
        compare(list(config), expected=[1, 2])

    def test_int(self):
        # not very useful, but documents .data as a public api
        config = ConfigNode(1)
        compare(config.data, 1)


class TestItemAccess(object):

    def test_there_dict(self):
        config = ConfigNode({'foo': 1})
        compare(config['foo'], expected=1)

    def test_there_list(self):
        config = ConfigNode([1])
        compare(config[0], expected=1)

    def test_not_there_dict(self):
        config = ConfigNode({})
        with ShouldRaise(KeyError('foo')):
            config['foo']

    def test_not_there_list(self):
        config = ConfigNode([])
        with ShouldRaise(IndexError('list index out of range')):
            config[0]

    def test_not_there_simple(self):
        config = ConfigNode(1)
        with ShouldRaise(KeyError('foo')):
            config['foo']

    def test_get(self):
        config = ConfigNode({'foo': 1})
        compare(config.get('foo'), expected=1)

    def test_get_default(self):
        config = ConfigNode({})
        compare(config.get('foo', 1), expected=1)

    def test_get_default_default(self):
        config = ConfigNode({})
        compare(config.get('foo'), expected=None)

    def test_get_list_number_key(self):
        config = ConfigNode([1])
        compare(config.get(0), expected=1)

    def test_get_list_string_key(self):
        config = ConfigNode([1])
        compare(config.get('foo'), expected=None)

    def test_simple_value(self):
        config = ConfigNode({})
        compare(config.get('foo', 1), expected=1)

    def test_dict_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_list_value(self):
        config = ConfigNode({'foo': [1]})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_get_wrapped_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_items_dict(self):
        config = ConfigNode({'foo': 1, 'bar': 2})
        compare(config.items(), expected=[('bar', 2), ('foo', 1)])

    def test_items_list(self):
        config = ConfigNode([])
        expected = "'list' object has no attribute 'items'"
        with ShouldRaise(AttributeError(expected)):
            tuple(config.items())

    def test_items_wrapped_value(self):
        config = ConfigNode({'foo': [1]})
        obj = tuple(config.items())[0][1]
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_set_item_dict(self):
        config = ConfigNode()
        config['foo'] = 1
        compare(config.data, expected={'foo': 1})

    def test_set_item_list(self):
        config = ConfigNode(['old'])
        config[0] = 'new'
        compare(config.data, expected=['new'])


class TestAttributeAccess(object):

    def test_there(self):
        config = ConfigNode({'foo': 1})
        compare(config.foo, expected=1)

    def test_not_there(self):
        config = ConfigNode({})
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_simple_value(self):
        config = ConfigNode({'foo': 1})
        compare(config.foo, expected=1)

    def test_dict_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config.foo
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_list_value(self):
        config = ConfigNode({'foo': [1]})
        obj = config.foo
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_name_clash_with_real_attrs(self):
        # method, data, __method
        config = ConfigNode({'get': 1})
        compare(config.get('get'), expected=1)

    def test_list_source(self):
        config = ConfigNode([1])
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_simple_source(self):
        config = ConfigNode(1)
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_set_attr_dict(self):
        config = ConfigNode()
        config.foo = 1
        compare(config.data, expected={'foo': 1})

    def test_set_attr_list(self):
        config = ConfigNode([])
        with ShouldRaise(TypeError):
            config.foo = 1
        compare(config.data, expected=[])


class TestOtherFunctionality(object):

    def test_iterate_over_list_of_dicts(self):
        node = ConfigNode([{'x': 1}])
        compare(tuple(node)[0], expected=ConfigNode({'x': 2}))

    def test_iterate_over_dict(self):
        node = ConfigNode({'x': 1, 'y': 2})
        compare(sorted(node), expected=['x', 'y'])

    def test_set_on_root(self):
        node = ConfigNode({'x': 1, 'y': 2})
        node.set(['a', 'b'])
        compare(node.data, expected=['a', 'b'])

    def test_repr(self):
        node = ConfigNode({'some long key': 'some\nvalue',
                            'another long key': 2,
                            'yet another long key': 3})
        compare(repr(node), expected=dedent("""\
            configurator.node.ConfigNode(
            {'another long key': 2,
             'some long key': 'some\\nvalue',
             'yet another long key': 3}
            )"""))


class TestNodeActions(object):

    def test_attr_dict(self):
        node = ConfigNode({'a': {'b': 1}})
        compare(node.a.data, expected={'b': 1}, strict=True)
        node.a.set('c')
        compare(node.data, expected={'a': 'c'}, strict=True)

    def test_item_dict(self):
        node = ConfigNode({'a': {'b': 1}})
        compare(node.a.data, expected={'b': 1}, strict=True)
        node['a'].set('c')
        compare(node.data, expected={'a': 'c'}, strict=True)

    def test_item_list(self):
        node = ConfigNode([['a', 'b'], ['c', 'd']])
        compare(node[0].data, expected=['a', 'b'], strict=True)
        node[0].set('e')
        compare(node.data, expected=['e', ['c', 'd']], strict=True)

    def test_items(self):
        node = ConfigNode({'a': [1], 'b': [2]})
        keys = []
        for key, child in node.items():
            keys.append(key)
            child.set(child[0]*2)
        compare(node.data, expected={'a': 2, 'b': 4}, strict=True)
        compare(keys, expected=['a', 'b'])

    def test_iter_list(self):
        node = ConfigNode([['a', 'b'], ['c', 'd']])
        for child in node:
            child.set(''.join(child.data))
        compare(node.data, expected=['ab', 'cd'], strict=True)

    def test_node_empty(self):
        node = ConfigNode()
        assert node.node() is node

    def test_node_path_does_nothing(self):
        node = ConfigNode()
        assert node.node(source) is node

    def test_node_dict_item(self):
        node = ConfigNode({'a': 1, 'b': 2})
        child = node.node('a')
        compare(child.data, expected=1)
        child.set(3)
        compare(node.data, expected={'a': 3, 'b': 2}, strict=True)

    def test_node_list_item(self):
        node = ConfigNode(['a', 'b'])
        child = node.node(0)
        compare(child.data, expected='a')
        child.set('c')
        compare(node.data, expected=['c', 'b'], strict=True)

    def test_node_not_there(self):
        node = ConfigNode()
        with ShouldRaise(NotPresent('a')):
            node.node('a.b.c')

    def test_node_not_there_create(self):
        root = ConfigNode()
        child = root.node('a.b.c', create=True)
        compare(root.data, expected={'a': {'b': {'c': {}}}})
        child.set(1)
        compare(root.data, expected={'a': {'b': {'c': 1}}})

    def test_path_no_text_attribute(self):
        node = ConfigNode({'a': {'b': {'c': '1'}}})
        with ShouldRaise(TypeError("invalid path: convert(source['a']['b']['c'], int)")):
            node.node(convert(source['a']['b']['c'], int))



# See license.txt for license details.
# Copyright (c) 2011-2014 Simplistix Ltd, 2016-2020 Chris Withers

import os

from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)

setup(
    name='configurator',
    version='2.2.0',
    author='Chris Withers',
    author_email='chris@withers.org',
    license='MIT',
    description=(
        "A library for building a configuration store "
        "from one or more layered configuration sources"
    ),
    long_description=open('README.rst').read(),
    url='https://github.com/Simplistix/configurator',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(exclude=["tests"]),
    zip_safe=False,
    include_package_data=True,
    extras_require=dict(
        yaml=['pyyaml'],
        toml=['toml'],
        test=[
            'jinja2',
            'mock',
            'pyfakefs',
            'pytest',
            'pytest-cov',
            'sybil',
            'testfixtures',
            'voluptuous',
        ],
        build=['sphinx', 'sphinx-rtd-theme', 'setuptools-git', 'twine', 'wheel']
    ),
)
